from typing import cast
from functools import reduce

from .structural import (
    PsAssignment,
    PsAstNode,
    PsBlock,
    PsEmptyLeafMixIn,
    PsConditional,
    PsDeclaration,
    PsExpression,
    PsLoop,
    PsStatement,
)
from .expressions import PsSymbolExpr, PsConstantExpr

from ..symbols import PsSymbol
from ..exceptions import PsInternalCompilerError


class UndefinedSymbolsCollector:
    """Collect undefined symbols.

    This class implements an AST visitor that collects all symbols that have been used
    in the AST without being defined prior to their usage.
    """

    def __call__(self, node: PsAstNode) -> set[PsSymbol]:
        """Returns all symbols that occur in the given AST without being defined prior to their usage."""
        return self.visit(node)

    def visit(self, node: PsAstNode) -> set[PsSymbol]:
        undefined_vars: set[PsSymbol] = set()

        match node:
            case PsExpression():
                return self.visit_expr(node)

            case PsStatement(expr):
                return self.visit_expr(expr)

            case PsAssignment(lhs, rhs):
                undefined_vars = self(lhs) | self(rhs)
                if isinstance(lhs, PsSymbolExpr):
                    undefined_vars.remove(lhs.symbol)
                return undefined_vars

            case PsBlock(statements):
                for stmt in statements[::-1]:
                    undefined_vars -= self.declared_variables(stmt)
                    undefined_vars |= self(stmt)

                return undefined_vars

            case PsLoop(ctr, start, stop, step, body):
                undefined_vars = self(start) | self(stop) | self(step) | self(body)
                undefined_vars.discard(ctr.symbol)
                return undefined_vars

            case PsConditional(cond, branch_true, branch_false):
                undefined_vars = self(cond) | self(branch_true)
                if branch_false is not None:
                    undefined_vars |= self(branch_false)
                return undefined_vars

            case PsEmptyLeafMixIn():
                return set()

            case unknown:
                raise PsInternalCompilerError(
                    f"Don't know how to collect undefined variables from {unknown}"
                )

    def visit_expr(self, expr: PsExpression) -> set[PsSymbol]:
        match expr:
            case PsSymbolExpr(symb):
                return {symb}
            case _:
                return reduce(
                    set.union,
                    (self.visit_expr(cast(PsExpression, c)) for c in expr.children),
                    set(),
                )

    def declared_variables(self, node: PsAstNode) -> set[PsSymbol]:
        """Returns the set of variables declared by the given node which are visible in the enclosing scope."""

        match node:
            case PsDeclaration():
                return {node.declared_symbol}

            case (
                PsAssignment()
                | PsBlock()
                | PsConditional()
                | PsExpression()
                | PsLoop()
                | PsStatement()
                | PsEmptyLeafMixIn()
            ):
                return set()

            case unknown:
                raise PsInternalCompilerError(
                    f"Don't know how to collect declared variables from {unknown}"
                )


def collect_undefined_symbols(node: PsAstNode) -> set[PsSymbol]:
    return UndefinedSymbolsCollector()(node)


def collect_required_headers(node: PsAstNode) -> set[str]:
    match node:
        case PsSymbolExpr(symb):
            return symb.get_dtype().required_headers
        case PsConstantExpr(cs):
            return cs.get_dtype().required_headers
        case _:
            return reduce(
                set.union, (collect_required_headers(c) for c in node.children), set()
            )
